"""
Handles publishing arrays, videos, and cameras.

CamCtrl handles sending and receiving commands to specific camera (or array/video) publishers
VideoHandlerThread updates the frames for the global displayer, since OpenCV can only update on the main thread
get_cam_ids gets the ids for all cameras that OpenCV can detect
pub_cam_thread continually publishes updates to arrays, videos, and cameras
np_cam simulates numpy arrays as OpenCV cameras
"""

from . import subscriber_dictionary
from .frame_updater import FrameUpdater
from .get_frame_ids import get_cam_ids
from .np_to_opencv import NpCam
from .frame_publishing import pub_cam_thread



"""Get camera IDs."""

import cv2

from typing import List


def get_cam_ids() -> List[int]:
    """Get all cameras that OpenCV can currently detect."""
    cam_list: List[int] = []

    while True:
        cam = cv2.VideoCapture(len(cam_list))
        if not cam.isOpened():
            break
        cam_list.append(len(cam_list))

    return cam_list



"""Commands to control the array displaying windows."""

from localpubsub import VariablePub, VariableSub

key_pub = VariablePub()
mouse_pub = VariablePub()
win_cmd_pub = VariablePub()


def quit(force_all_read=True):
    """Quit the main loop displaying all the windows."""
    win_cmd_pub.publish("quit", force_all_read=force_all_read)


def win_cmd_sub() -> VariableSub:
    """Get a subscriber to the main window loop."""
    return win_cmd_pub.make_sub()



"""Transform functions."""

import numpy as np
import cv2


def transform_about_center(
    arr, scale_multiplier=(1, 1), rotation_degrees=0, translation=(0, 0), skew=(0, 0)
):
    """
    Transform an image about its center.

    :param arr: numpy image to be transformed
    :param scale_multiplier: grow/shrink in x and y
    :param rotation_degrees: degrees to rotate, from 0 to 360
    :param translation: pixels to translate the image
    :param skew: mimics rotation along screen axes. In degrees. 90 degrees should give a line.
    :return: transformed numpy image
    """
    center_scale_xform = np.eye(3)
    center_scale_xform[0, 0] = scale_multiplier[1]
    center_scale_xform[1, 1] = scale_multiplier[0]
    center_scale_xform[0:2, -1] = [arr.shape[1] // 2, arr.shape[0] // 2]

    rotation_xform = np.eye(3)

    theta = np.radians(rotation_degrees)
    c, s = np.cos(theta), np.sin(theta)
    R = np.array(((c, -s), (s, c)))
    rotation_xform[0:2, 0:2] = R
    skew = np.radians(skew)
    skew = np.tan(skew)
    rotation_xform[-1, 0:2] = [skew[1] / arr.shape[1], skew[0] / arr.shape[0]]

    translation_skew_xform = np.eye(3)
    translation_skew_xform[0:2, -1] = [
        (-arr.shape[1] - translation[1]) // 2,
        (-arr.shape[0] - translation[0]) // 2,
    ]

    full_xform = center_scale_xform @ rotation_xform @ translation_skew_xform
    xformd_arr = cv2.warpPerspective(
        arr, full_xform, tuple(reversed(arr.shape[:2])), flags=0
    )
    return xformd_arr



"""Functions needed to deal with OpenCV."""

import weakref


class WeakMethod(weakref.WeakMethod):
    """Pass any method to OpenCV without it keeping a reference forever."""

    def __call__(self, *args, **kwargs):
        """Call the actual method this object was made with."""
        obj = super().__call__()
        func = self._func_ref()
        if obj is None or func is None:
            return None
        meth = self._meth_type(func, obj)
        meth(*args, **kwargs)



"""Allow OpenCV to handle zmq subscriber addresses as input."""

import cv2
import zmq
from tensorcom.tenbin import decode_buffer  # type: ignore


class ZmqCam(object):
    """Add OpenCV camera controls to a numpy array."""

    def __init__(self, img):
        """Create a fake camera for OpenCV based on the initial array."""
        assert isinstance(img, str)
        s = img.split('#')
        self.__ctx = zmq.Context()
        self.__addr = s[0]
        self.__sub = self.__ctx.socket(zmq.SUB)
        if len(s) > 1:
            self.__topic = bytes(s[1], 'ascii')
            self.__sub.setsockopt(zmq.SUBSCRIBE, self.__topic)
        else:
            self.__topic = b""
        self.__sub.connect(self.__addr)

        self.__is_opened = True

    def set(self, *args, **kwargs):
        """Set CAP_PROP_FRAME_WIDTH or CAP_PROP_FRAME_HEIGHT to scale a numpy array to that size."""
        pass

    @staticmethod
    def get(*args, **kwargs):
        """Get OpenCV args. Currently only a fake CAP_PROP_FRAME_COUNT to fix detecting video ends."""
        if args[0] == cv2.CAP_PROP_FRAME_COUNT:
            return float("inf")

    def read(self):
        """Read back the numpy array in standard "did it work", "the array", OpenCV format."""
        r = self.__sub.recv_multipart()
        arrs = [decode_buffer(ri) for ri in r[1:]]
        return True, arrs

    def isOpened(self):  # NOSONAR
        """Hack to tell OpenCV we're opened until we call release."""
        return self.__is_opened

    def release(self):
        """Let OpenCV know we're finished."""
        self.__is_opened = False



"""Allow OpenCV to handle numpy arrays as input."""

import numpy as np
import cv2


class NpCam(object):
    """Add OpenCV camera controls to a numpy array."""

    def __init__(self, img):
        """Create a fake camera for OpenCV based on the initial array."""
        assert isinstance(img, np.ndarray)
        self.__img = img
        self.__is_opened = True
        if len(img.shape) > 0:
            self.__height = self.__img.shape[0]
            if len(self.__img.shape) > 1:
                self.__width = self.__img.shape[1]
            else:
                self.__width = self.__height
        else:
            self.__width = self.__height = 1
        self.__ratio = float(self.__width) / self.__height

        self.__wait_for_ratio = False

    def __handler_ratio(self):
        if self.__width <= 0 or not isinstance(self.__width, int):
            self.__width = int(self.__ratio * self.__height)
        elif self.__height <= 0 or not isinstance(self.__height, int):
            self.__height = int(self.__width / self.__ratio)
        if self.__width > 0 and self.__height > 0:
            self.__img = cv2.resize(self.__img, (self.__width, self.__height))

    def set(self, *args, **kwargs):
        """Set CAP_PROP_FRAME_WIDTH or CAP_PROP_FRAME_HEIGHT to scale a numpy array to that size."""
        if args[0] in [cv2.CAP_PROP_FRAME_WIDTH, cv2.CAP_PROP_FRAME_HEIGHT]:
            self.__wait_for_ratio = not self.__wait_for_ratio
            if args[0] == cv2.CAP_PROP_FRAME_WIDTH:
                self.__width = args[1]
            else:
                self.__height = args[1]
            if not self.__wait_for_ratio:
                self.__handler_ratio()

    @staticmethod
    def get(*args, **kwargs):
        """Get OpenCV args. Currently only a fake CAP_PROP_FRAME_COUNT to fix detecting video ends."""
        if args[0] == cv2.CAP_PROP_FRAME_COUNT:
            return float("inf")

    def read(self):
        """Read back the numpy array in standard "did it work", "the array", OpenCV format."""
        return True, self.__img

    def isOpened(self):  # NOSONAR
        """Hack to tell OpenCV we're opened until we call release."""
        return self.__is_opened

    def release(self):
        """Let OpenCV know we're finished."""
        self.__is_opened = False



"""Standard callbacks to use on incoming frames."""

from displayarray.window import window_commands
import numpy as np

from typing import Union


def global_cv_display_callback(frame: np.ndarray, cam_id: Union[int, str]):
    """
    Send frames to the global frame dictionary.

    :param frame: The video or image frame
    :type frame: np.ndarray
    :param cam_id: The video or image source
    :type cam_id: Union[int, str]
    """
    from displayarray.window import SubscriberWindows

    SubscriberWindows.FRAME_DICT[str(cam_id)] = frame


class function_display_callback(object):  # NOSONAR
    """
    Used for running arbitrary functions on pixels.

    .. code-block:: python

      >>> import random
      >>> from displayarray.frame import FrameUpdater
      >>> img = np.zeros((300, 300, 3))
      >>> def fun(array, coords, finished):
      ...     r,g,b = random.random()/20.0, random.random()/20.0, random.random()/20.0
      ...     array[coords[0:2]] = (array[coords[0:2]] + [r,g,b])%1.0
      >>> FrameUpdater(video_source=img, callbacks=function_display_callback(fun)).display()

    :param display_function: a function to run on the input image.
    :param finish_function: a function to run on the input image when the other function finishes.
    """

    def __init__(self, display_function, finish_function=None):
        """Run display_function on frames."""
        self.looping = True
        self.first_call = True

        def _run_finisher(self, frame, finished, *args, **kwargs):
            if not callable(finish_function):
                window_commands.quit()
            else:
                finished = finish_function(frame, Ellipsis, finished, *args, **kwargs)
                if finished:
                    window_commands.quit()

        def _display_internal(self, frame, *args, **kwargs):
            finished = True
            if self.first_call:
                # return to display initial frame
                self.first_call = False
                return
            if self.looping:
                it = np.nditer(frame, flags=["multi_index"])
                while not it.finished:
                    x, y, c = it.multi_index
                    finished = display_function(
                        frame, (x, y, c), finished, *args, **kwargs
                    )
                    it.iternext()
            if finished:
                self.looping = False
                _run_finisher(self, frame, finished, *args, **kwargs)

        self.inner_function = _display_internal

    def __call__(self, *args, **kwargs):
        """Call the function "function_display_callback" was set up with."""
        return self.inner_function(self, *args, **kwargs)



# -*- coding: utf-8 -*-

# DO NOT EDIT THIS FILE!
# This file has been autogenerated by dephell <3
# https://github.com/dephell/dephell

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup


import os.path

readme = ""
here = os.path.abspath(os.path.dirname(__file__))
readme_path = os.path.join(here, "README.rst")
if os.path.exists(readme_path):
    with open(readme_path, "rb") as stream:
        readme = stream.read().decode("utf8")


setup(
    long_description=readme,
    name="displayarray",
    version="1.3.1",
    description="Tool for displaying numpy arrays.",
    python_requires="==3.*,>=3.7.0",
    project_urls={"repository": "https://github.com/simleek/displayarray"},
    author="SimLeek",
    author_email="simulator.leek@gmail.com",
    license="MIT",
    entry_points={"console_scripts": ["displayarray = displayarray.__main__:main"]},
    packages=[
        "displayarray",
        "displayarray.effects",
        "displayarray.frame",
        "displayarray.window",
    ],
    package_dir={"": "."},
    package_data={},
    install_requires=[
        "docopt==0.6.2",
        "localpubsub==0.0.4",
        "numpy>=1.14.5",
        "opencv-python==4.*,>=4.0.0",
        "pyzmq>=22.0.3",
        "tensorcom"
    ],
    extras_require={
        # pypi doesn't allow direct dependencies for security reasons,
        # even though I could install a lot of viruses just from this setup.py
        # "linux": ["PyV4L2Cam @ git+https://github.com/SimLeek/PyV4L2Cam"],
        "dev": [
            "black==18.*,>=18.3.0.a0",
            "coverage==4.*,>=4.5.0",
            "mock==3.*,>=3.0.0",
            "mypy==0.*,>=0.740.0",
            "pydocstyle==4.*,>=4.0.0",
            "pytest==5.2.1",
            "sphinx==2.*,>=2.2.0",
            "tox==3.*,>=3.14.0",
            "tox-gh-actions==0.*,>=0.3.0",
            "typing==3.7.4.1",
            "wheel==0.*,>=0.30.0"
        ],
    },
)



"""Reduce many color images to the three colors that your eyeballs can see."""

import numpy as np
from ..input import mouse_loop
import cv2

from typing import Iterable


class SelectChannels(object):
    """
    Select channels to display from an array with too many colors.

    :param selected_channels: the list of channels to display.
    """

    def __init__(self, selected_channels: Iterable[int] = None):
        """Select which channels from the input array to display in the output."""
        if selected_channels is None:
            selected_channels = [0, 0, 0]
        self.selected_channels = selected_channels
        self.mouse_control = None
        self.mouse_print_channels = False
        self.num_input_channels = None

    def __call__(self, arr):
        """Run the channel selector."""
        if isinstance(arr, list):
            ars = []
            for a in arr:
                ars.append(self.__call__(a))
            return ars
        else:
            self.num_input_channels = arr.shape[-1]
            out_arr = [
                arr[..., min(max(0, x), arr.shape[-1] - 1)] for x in self.selected_channels
            ]
            out_arr = np.stack(out_arr, axis=-1)
            return out_arr

    def enable_mouse_control(self):
        """
        Enable mouse control.

        Alt+Scroll to increase/decrease channel 2.
        Shift+Scroll to increase/decrease channel 1.
        Ctrl+scroll to increase/decrease channel 0.
        """

        @mouse_loop
        def m_loop(me):
            if me.event == cv2.EVENT_MOUSEWHEEL:
                if me.flags & cv2.EVENT_FLAG_CTRLKEY:
                    if me.flags > 0:
                        self.selected_channels[0] += 1
                        self.selected_channels[0] = min(
                            self.selected_channels[0], self.num_input_channels - 1
                        )
                    else:
                        self.selected_channels[0] -= 1
                        self.selected_channels[0] = max(self.selected_channels[0], 0)
                    if self.mouse_print_channels:
                        print(f"Channel 0 now maps to {self.selected_channels[0]}.")
                elif me.flags & cv2.EVENT_FLAG_SHIFTKEY:
                    if me.flags > 0:
                        self.selected_channels[1] += 1
                        self.selected_channels[1] = min(
                            self.selected_channels[1], self.num_input_channels - 1
                        )
                    else:
                        self.selected_channels[1] -= 1
                        self.selected_channels[1] = max(self.selected_channels[1], 0)
                    if self.mouse_print_channels:
                        print(f"Channel 1 now maps to {self.selected_channels[1]}.")
                elif me.flags & cv2.EVENT_FLAG_ALTKEY:
                    if me.flags > 0:
                        self.selected_channels[2] += 1
                        self.selected_channels[2] = min(
                            self.selected_channels[2], self.num_input_channels - 1
                        )
                    else:
                        self.selected_channels[2] -= 1
                        self.selected_channels[2] = max(self.selected_channels[2], 0)
                    if self.mouse_print_channels:
                        print(f"Channel 2 now maps to {self.selected_channels[2]}.")

        self.mouse_control = m_loop



"""Crop any n-dimensional array."""

import numpy as np
from ..input import mouse_loop


class Crop(object):
    """A callback class that will return the input array cropped to the output size. N-dimensional."""

    def __init__(self, output_size=(64, 64, 3), center=None):
        """
        Create the cropping callback class.

        :param output_size: Specified the size the input should be cropped to. Can be redefined later.
        :param center: Specifies the center on the input array to take the crop out of.
        """
        self._output_size = None
        self._center = np.asarray([o // 2 for o in output_size])
        self.odd_center = None
        self.mouse_control = None
        self.input_size = None

        self.center = center
        self.output_size = output_size

    @property
    def output_size(self):
        """Get the output size."""
        return self._output_size

    @output_size.setter
    def output_size(self, set):
        """Set the output size."""
        self._output_size = set
        if self._output_size is not None:
            self._output_size = np.asarray(set)

    @property
    def center(self):
        """Get the center."""
        return self._center

    @center.setter
    def center(self, set):
        """Set the center. Guarded so that colors need not be set."""
        if set is not None:
            for x in range(len(set)):
                self._center[x] = set[x]

    def __call__(self, arr):
        """Crop the input array to the specified output size. output is centered on self.center point on input."""
        self.input_size = arr.shape
        if self.center is None:
            self.center = [int(arr.shape[x]) // 2 for x in range(arr.ndim)]
        self.odd_out = np.array(
            [self.output_size[x] % 2 for x in range(len(self.output_size))]
        )
        self.odd_center = np.array(
            [self.center[x] % 2 for x in range(len(self.center))]
        )

        center = self.center.copy()  # stop opencv from thread breaking us
        top_left_get = [
            min(max(0, center[x] - self.output_size[x] // 2), arr.shape[x] - 1)
            for x in range(arr.ndim)
        ]
        bottom_right_get = [
            min(
                max(0, center[x] + self.output_size[x] // 2 + self.odd_out[x]),
                arr.shape[x],
            )
            for x in range(arr.ndim)
        ]

        top_left_put = [
            min(
                max(
                    0,
                    -(bottom_right_get[x] - center[x] - self.output_size[x] // 2)
                    + self.odd_out[x],
                ),
                self.output_size[x],
            )
            for x in range(arr.ndim)
        ]
        bottom_right_put = [
            min(
                max(0, top_left_put[x] + (bottom_right_get[x] - top_left_get[x])),
                self.output_size[x],
            )
            for x in range(arr.ndim)
        ]
        get_slices = [slice(x1, x2) for x1, x2 in zip(top_left_get, bottom_right_get)]
        get_slices = tuple(get_slices)
        put_slices = [slice(x1, x2) for x1, x2 in zip(top_left_put, bottom_right_put)]
        put_slices = tuple(put_slices)
        out_array = np.zeros(self.output_size)
        out_array[put_slices] = arr[get_slices]
        return out_array.astype(arr.dtype)

    def enable_mouse_control(self):
        """Move the mouse to move where the crop is from on the original image."""

        @mouse_loop
        def m_loop(me):
            if self.center is None:
                self.center = [0, 0, 1]
            self.center[:] = [
                int(me.y / self.output_size[0] * self.input_size[0]),
                int(me.x / self.output_size[1] * self.input_size[1]),
                1,
            ]

        self.mouse_control = m_loop
        return self



"""
DisplayArray.

Display NumPy arrays.

Usage:
  displayarray (-w <webcam-number> | -v <video-filename> | -t <topic-name>[,dtype])... [-m <msg-backend>]
  displayarray -h
  displayarray --version


Options:
  -h, --help                                           Show this help text.
  --version                                            Show version number.
  -w <webcam-number>, --webcam=<webcam-number>         Display video from a webcam.
  -v <video-filename>, --video=<video-filename>        Display frames from a video file.
  -t <topic-name>, --topic=<topic-name>                Display frames from a topic using the chosen message broker.
  -m <msg-backend>, --message-backend <msg-backend>    Choose message broker backend. [Default: ROS]
                                                       Currently supported: ROS, ZeroMQ
  --ros                                                Use ROS as the backend message broker.
  --zeromq                                             Use ZeroMQ as the backend message broker.
"""

from docopt import docopt
import asyncio


def main(argv=None):
    """Process command line arguments."""
    arguments = docopt(__doc__, argv=argv)
    if arguments["--version"]:
        from displayarray import __version__

        print(f"DisplayArray V{__version__}")
        return
    from displayarray import display

    vids = [int(w) for w in arguments["--webcam"]] + arguments["--video"]
    v_disps = None
    if vids:
        v_disps = display(*vids, blocking=False)
    from displayarray.frame.frame_updater import read_updates_ros, read_updates_zero_mq

    topics = arguments["--topic"]
    topics_split = [t.split(",") for t in topics]
    d = display()

    async def msg_recv():
        nonlocal d
        while d:
            if arguments["--message-backend"] == "ROS":
                async for v_name, frame in read_updates_ros(
                    [t for t, d in topics_split], [d for t, d in topics_split]
                ):
                    d.update(arr=frame, id=v_name)
            if arguments["--message-backend"] == "ZeroMQ":
                async for v_name, frame in read_updates_zero_mq(
                    *[bytes(t, encoding="ascii") for t in topics]
                ):
                    d.update(arr=frame, id=v_name)

    async def update_vids():
        while v_disps:
            if v_disps:
                v_disps.update()
                await asyncio.sleep(0)

    async def runner():
        await asyncio.wait([msg_recv(), update_vids()])

    loop = asyncio.get_event_loop()
    loop.run_until_complete(runner())
    loop.close()


if __name__ == "__main__":
    main()



"""Publish frames so any function within this program can find them."""

import asyncio
import sys
import threading
import time
import warnings

import cv2

using_pyv4l2cam = False
try:
    if sys.platform == "linux":
        from PyV4L2Cam.camera import Camera as pyv4lcamera  # type: ignore
        from PyV4L2Cam.controls import ControlIDs as pyv4lcontrolids  # type: ignore
        from PyV4L2Cam import convert_mjpeg, convert_rgb24  # type: ignore
        from PyV4L2Cam.get_camera import get_camera_by_bus_info, get_camera_by_string  # type: ignore

        using_pyv4l2cam = True
except ImportError:
    warnings.warn("Could not import PyV4L2Cam on linux. Camera capture will be slow.")
    warnings.warn(
        "To install, run: pip install git+https://github.com/simleek/PyV4L2Cam.git"
    )

import numpy as np

from displayarray.frame import subscriber_dictionary
from .np_to_opencv import NpCam
from .zmq_to_opencv import ZmqCam
from displayarray._uid import uid_for_source

from typing import Union, Tuple, Optional, Dict, Any, List, Callable

FrameCallable = Callable[[np.ndarray], Optional[np.ndarray]]


def pub_cam_loop_pyv4l2(
    cam_id: Union[int, str, np.ndarray],
    request_size: Tuple[int, int] = (-1, -1),
    mjpg: bool = True,
    fps_limit: float = float("inf"),
):
    """
    Publish whichever camera you select to CVCams.<cam_id>.Vid, using v4l2 instead of opencv.

    You can send a quit command 'quit' to CVCams.<cam_id>.Cmd
    Status information, such as failure to open, will be posted to CVCams.<cam_id>.Status

    :param mjpg: Selects mjpeg transferring, which most cameras seem to support, so speed isn't limited
    :param fps_limit: Limits the frames per second.
    :param cam_id: An integer representing which webcam to use, or a string representing a video file.
    :param request_size: A tuple with width, then height, to request the video size.
    :return: True if loop ended normally, False if it failed somehow.
    """
    name = uid_for_source(cam_id)

    if isinstance(cam_id, (int, str)):
        if isinstance(cam_id, int):
            cam: pyv4lcamera = pyv4lcamera(  # type: ignore
                f"/dev/video{cam_id}", *request_size
            )
        else:
            if "usb" in cam_id:
                cam = get_camera_by_bus_info(cam_id, *request_size)  # type: ignore
            else:
                cam = get_camera_by_string(cam_id, *request_size)  # type: ignore
    else:
        raise TypeError(
            "Only strings or ints representing cameras are supported with v4l2."
        )

    subscriber_dictionary.register_cam(name, cam)

    sub = subscriber_dictionary.cam_cmd_sub(name)
    sub.return_on_no_data = ""
    msg = ""

    if mjpg and cam.pixel_format != "MJPEG":
        warnings.warn("Camera does not support high speed.")

    now = time.time()
    while msg != "quit":
        time.sleep(1.0 / (fps_limit - (time.time() - now)))
        now = time.time()
        frame_bytes = cam.get_frame()  # type: bytes

        if cam.pixel_format == "MJPEG":
            nd_frame = convert_mjpeg(frame_bytes)  # type: ignore
        elif cam.pixel_format == "RGB24":
            nd_frame = convert_rgb24(frame_bytes, cam.width, cam.height)  # type: ignore
        else:
            raise NotImplementedError(f"{cam.pixel_format} format not supported.")

        if nd_frame is not None:
            subscriber_dictionary.CV_CAMS_DICT[name].frame_pub.publish(nd_frame)
        else:
            cam.close()
            subscriber_dictionary.CV_CAMS_DICT[name].status_pub.publish("failed")
            return False

        msg = sub.get()
    sub.release()

    cam.close()
    return True


def pub_cam_loop_opencv(
    cam_id: Union[int, str, np.ndarray],
    request_size: Tuple[int, int] = (-1, -1),
    mjpg: bool = True,
    fps_limit: float = float("inf"),
    extra: Optional[List[Tuple[int, int]]] = None,
) -> bool:
    """
    Publish whichever camera you select to CVCams.<cam_id>.Vid.

    You can send a quit command 'quit' to CVCams.<cam_id>.Cmd
    Status information, such as failure to open, will be posted to CVCams.<cam_id>.Status

    :param mjpg: Selects mjpeg transferring, which most cameras seem to support, so speed isn't limited
    :param fps_limit: Limits the frames per second.
    :param cam_id: An integer representing which webcam to use, or a string representing a video file.
    :param request_size: A tuple with width, then height, to request the video size.
    :return: True if loop ended normally, False if it failed somehow.
    """
    name = uid_for_source(cam_id)

    cam: Union[NpCam, ZmqCam, cv2.VideoCapture]
    if isinstance(cam_id, (int, str)):
        if isinstance(cam_id, str) and cam_id.startswith('tcp'):
            cam = ZmqCam(cam_id)
        else:
            cam = cv2.VideoCapture(cam_id)
    elif isinstance(cam_id, (np.ndarray)):
        cam = NpCam(cam_id)
    else:
        raise TypeError(
            "Only strings or ints representing cameras, or numpy arrays representing pictures supported."
        )

    subscriber_dictionary.register_cam(name, cam)

    frame_counter = 0

    sub = subscriber_dictionary.cam_cmd_sub(name)
    sub.return_on_no_data = ""
    msg = ""
    if mjpg:
        try:
            cam.set(cv2.CAP_PROP_FOURCC, cv2.CAP_OPENCV_MJPEG)
        except AttributeError:
            warnings.warn("Please update OpenCV")

    cam.set(cv2.CAP_PROP_FRAME_WIDTH, request_size[0])
    cam.set(cv2.CAP_PROP_FRAME_HEIGHT, request_size[1])

    if not cam.isOpened():
        subscriber_dictionary.CV_CAMS_DICT[name].status_pub.publish("failed")
        return False
    now = time.time()
    while msg != "quit":
        time.sleep(1.0 / (fps_limit - (time.time() - now)))
        now = time.time()
        (ret, frame) = cam.read()  # type: Tuple[bool, np.ndarray ]
        if ret is False or not isinstance(frame, (np.ndarray, list)):
            cam.release()
            subscriber_dictionary.CV_CAMS_DICT[name].status_pub.publish("failed")
            return False
        if cam.get(cv2.CAP_PROP_FRAME_COUNT) > 0:
            frame_counter += 1
            if frame_counter >= cam.get(cv2.CAP_PROP_FRAME_COUNT):
                frame_counter = 0
                cam = cv2.VideoCapture(cam_id)
        try:
            subscriber_dictionary.CV_CAMS_DICT[name].frame_pub.publish(frame)
        except KeyError:  # we got deleted. Time to exit.
            cam.release()
            return False
        msg = sub.get()
    sub.release()

    cam.release()
    return True


uid_dict: Dict[str, threading.Thread] = {}


def pub_cam_thread(
    cam_id: Union[int, str],
    request_ize: Tuple[int, int] = (-1, -1),
    mjpg: bool = True,
    fps_limit: float = float("inf"),
    force_backend="",
) -> threading.Thread:
    """Run pub_cam_loop in a new thread. Starts on creation."""

    name = uid_for_source(cam_id)
    t = None
    if name in uid_dict.keys():
        t = uid_dict[name]
    if t is None or not t.is_alive():  # Enables reopening cameras
        if "cv" in force_backend.lower():
            pub_cam_loop = pub_cam_loop_opencv
        elif (
            sys.platform == "linux"
            and using_pyv4l2cam
            and (
                isinstance(cam_id, int)
                or (
                    isinstance(cam_id, str)
                    and any(["/dev/video" in cam_id, "usb" in cam_id])
                )
            )
        ) or "v4l2" in force_backend.lower():
            pub_cam_loop = pub_cam_loop_pyv4l2  # type: ignore
        else:
            pub_cam_loop = pub_cam_loop_opencv

        t = threading.Thread(
            target=pub_cam_loop, args=(cam_id, request_ize, mjpg, fps_limit)
        )
        uid_dict[name] = t
        t.start()
    return t
